// sylvan_analyzer.nut

// -- Sorter--------------------------------------------------------------
class Sorter {
  static function standard_compare( x, y ) {
    if ( x.board_type != y.board_type ) {
      return x.board_type - y.board_type;
    }
    if ( x.board_type == Board.Type.Sideboard ) {
      // TCh{[h͕ёւȂǁuSideboardStartv͐擪
      if ( x.card.type == Card.Type.SideboardStart || y.card.type == Card.Type.SideboardStart ) {
        return x.card.type - y.card.type;
      }
      return 0;
    }
    return Card.compare( x.card, y.card );
  }

  static function type_compare( x, y ) {
    if ( x.board_type != y.board_type ) {
      return x.board_type - y.board_type;
    }
    return x.card.type - y.card.type;
  }

  static function compare_same_board_normal_only( x, y, compare ) {
    if ( x.board_type != y.board_type ) {
      return x.board_type - y.board_type;
    }
    if ( x.board_type == Board.Type.Sideboard ) {
      // uSideboardStartv͐擪
      if ( x.card.type == Card.Type.SideboardStart || y.card.type == Card.Type.SideboardStart ) {
        return x.card.type - y.card.type;
      }
    }
    if ( x.card.is_normal() && y.card.is_normal() ) {
      return compare( x, y );
    }
    if ( x.card.is_normal() ) {
      return -1;
    }
    if ( y.card.is_normal() ) {
      return 1;
    }
    return x.card.type - y.card.type;
  }

  constructor( name, compare_for_sort ) {
    this.name = name;
    this.compare_for_sort = compare_for_sort;
  }

  name = null;
  compare_for_sort = null;
}

// -- RemainSorter -----
class RemainSorter extends Sorter {
  constructor() {
    this.name = "ёւ";
    this.compare_for_sort = function ( x, y ) { return 0; };
  }
}

// -- StandardSorter -----
class StandardSorter extends Sorter {
  constructor() {
    this.name = "W";
    this.compare_for_sort = Sorter.standard_compare;
  }
}

// -- TypeSorter -----
class TypeSorter extends Sorter {
  constructor() {
    this.name = "^Cv";
    this.compare_for_sort = Sorter.type_compare;
  }
}

// -- DeckSorter -----
class DeckSorter extends Sorter {
  constructor( deck ) {
    this.name = deck.file_name;
    this.deck = deck;
  }

  function compare_for_sort( x, y ) {
    return Sorter.compare_same_board_normal_only( x, y, function( x, y ) {
      local board = this.deck.get_board_type_of( x.board_type );
      return board.get_count_of( y.card ) - board.get_count_of( x.card );
    }.bindenv( this ) );
  }

  deck = null;
}

// -- TotalSorter -----
class TotalSorter extends Sorter {
  constructor() {
    this.name = "vl";
  }

  function compare_for_sort( x, y ) {
    return Sorter.compare_same_board_normal_only( x, y, function( x, y ) {
      return ::Organizer.get_sum_of_decks_card_count_by_entry( y ) - ::Organizer.get_sum_of_decks_card_count_by_entry( x );
    } );
  }
}

// -- CardCountSorter -----
class CardCountSorter extends Sorter {
  constructor( count ) {
    this.name = count.tostring() + "σfbL";
    this.count = count;
  }

  function compare_for_sort( x, y ) {
    return Sorter.compare_same_board_normal_only( x, y, function( x, y ) {
      local tmp = function( entry ) {
        return ::Organizer.get_sum_of_decks_board( entry.board_type, function ( board ) {
          return board.get_count_of( entry.card ) == this.count ? 1 : 0;
        }.bindenv( this ) );
      }
      return tmp( y ) - tmp( x );
    }.bindenv( this ) );
  }

  count = null;
}


// -- Processor ----------------------------------------------------------
class Processor {
  constructor( name, process_function, sorter ) {
    this.name = name;
    this.process_function = process_function;
    this.sorter = sorter;
  }

  name = null;
  process_function = function( entry ) { return null; };
  sorter = null;
}


// -- NameProcessor -----
class NameProcessor extends Processor {
  constructor() {
    this.name = "Oipj";
    this.sorter = ::StandardSorter();
  }

  function process_function( entry ) {
    return entry.card.name;
  }
}

// -- NameJapaneseProcessor -----
class NameJapaneseProcessor extends Processor {
  constructor() {
    this.name = "Oi{j";
    this.sorter = ::StandardSorter();
  }

  function process_function( entry ) {
    return entry.card.name_japanese;
  }
}

// -- BoardTypeProcessor -----
class BoardTypeProcessor extends Processor {
  constructor() {
    this.name = "{[h";
    this.sorter = ::StandardSorter();
  }

  function process_function( entry ) {
    switch( entry.board_type ) {
    case Board.Type.Mainboard:
      return "C";

    case Board.Type.Sideboard:
      return "TCh";

    case Board.Type.Hybrid:
      return "";
    }
    return null;
  }
}

// -- DeckProcessor -----
class DeckProcessor extends Processor {
  constructor( deck ) {
    this.deck = deck;
    this.name = deck.file_name;
    this.sorter = ::DeckSorter( deck );
  }

  function process_function( entry ) {
    local board = deck.get_board_type_of( entry.board_type );

    if ( entry.card.is_normal() ) {
      return board.get_count_of( entry.card );
    }
    if ( entry instanceof TotalEntry ) {
      return entry.count_total( board )
    }
    return null;
  }

  deck = null;
}

// -- TotalProcessor -----
class TotalProcessor extends Processor {
  constructor() {
    this.name = "v";
    this.sorter = ::TotalSorter();
  }

  function process_function( entry ) {
    if ( entry.card.is_normal() ) {
      return ::Organizer.get_sum_of_decks_card_count_by_entry( entry );
    }
    if ( entry instanceof TotalEntry ) {
      return ::Organizer.get_sum_of_decks_board( entry.board_type, @( board ) entry.count_total( board ) );
    }
    return null;
  }
}

// -- AverageProcessor -----
class AverageProcessor extends TotalProcessor {
  constructor() {
    this.name = "";
    this.sorter = ::TotalSorter();
  }

  function process_function( entry ) {
    local result = base.process_function( entry );
    if ( typeof( result ) == typeof( 0 ) ) {
      if ( ::Organizer.decks.len() == 0 ) {
        return 0.0;
      }
      return result.tofloat() / ::Organizer.decks.len().tofloat();
    }
    return result;
  }
}

// -- CardCountProcessor -----
class CardCountProcessor extends Processor {
  constructor( count ) {
    this.count = count
    this.name = count.tostring() + "σfbL";
    this.sorter = ::CardCountSorter( count );
  }

  function process_function( entry ) {
    if ( entry.card.is_normal() ) {
      return ::Organizer.get_sum_of_decks_board( entry.board_type, function( board ) {
        return board.get_count_of( entry.card ) == this.count ? 1 : 0;
      }.bindenv( this ) );
    }
    return null;
  }

  count = null;
}


// -- CardCountPercentageProcessor -----
class CardCountPercentageProcessor extends CardCountProcessor {
  constructor( count ) {
    base.constructor( count );
    this.name = count.tostring() + "σfbL";
    this.sorter = ::CardCountSorter( count );
  }

  function process_function( entry ) {
    local result = base.process_function( entry );
    if ( typeof( result ) == typeof( 0 ) ) {
      if ( ::Organizer.decks.len() == 0 ) {
        return "0.0%";
      }
      local tmp = result.tofloat() / ::Organizer.decks.len().tofloat();
      return ::format( "%.1f", tmp * 100.0 ) + "%";
    }
    return result;
  }
}



// -- TotalEntry ---------------------------------------------------------
class TotalEntry extends SpecialEntry {
  // constructor ͈ϏŔC

  // ( Board ̃CX^X ) ߂l integer ̊֐
  function count_total( board ) {
    return 0;
  }
}

// -- CardCountTotalEntry -----
class CardCountTotalEntry extends TotalEntry {
  // constructor ͈ϏŔC

  function count_total( board ) {
    return board.count_total_of( (@( card ) this.should_count( card )).bindenv( this ) );
  }

  //  card ̖vlɉǂԂ
  function should_count( card ) {
    return false;
  }
}

// -- TypeTotalEntry -----
class TypeTotalEntry extends CardCountTotalEntry {
  constructor( self_type, total_type, name, name_japanese, board_type ) {
    base.constructor( self_type, name, name_japanese, board_type );
    this.total_type = total_type
  }

  function should_count( card ) {
    return card.type == this.total_type;
  }

  total_type = null;
}

// -- InstantSorceryTotalEntry -----
class InstantSorceryTotalEntry extends CardCountTotalEntry {
  constructor( board_type ) {
    base.constructor( Card.Type.InstantSorceryTotal, "Instant Socery Total", "CX^g\[T[v", board_type );
  }

  function should_count( card ) {
    return card.is_instant_sorcery();
  }
}

// -- SpellTotalEntry -----
class SpellTotalEntry extends CardCountTotalEntry {
  constructor( board_type ) {
    base.constructor( Card.Type.SpellTotal, "Spell Total", "v", board_type );
  }

  function should_count( card ) {
    return card.is_spell();
  }
}

// -- AllTotalEntry -----
class AllTotalEntry extends CardCountTotalEntry {
  constructor( board_type ) {
    base.constructor( Card.Type.MainTotal, "All Total", "Sv", board_type );
  }

  function should_count( card ) {
    return true;
  }
}

// -- SideboardStartEntry -----
class SideboardStartEntry extends SpecialEntry {
  constructor() {
    base.constructor( ::Card.Type.SideboardStart, "Sideboard", "TCh{[h", Board.Type.Sideboard );
  }

  function case_of_cannot_process( processor ) {
    return "---";
  }
}


// -- Organizer ⑫ -----------------------------------------------------
::Organizer.get_sum_of_decks <- function( each_deck_function )
{
  local sum = 0;
  foreach ( deck in this.decks ) {
    sum += each_deck_function( deck );
  }
  return sum;
}

::Organizer.get_sum_of_decks_board <- function( board_type, each_board_function )
{
  return this.get_sum_of_decks( function( deck ) {
    return each_board_function( deck.get_board_type_of( board_type ) );
  } );
}

::Organizer.get_sum_of_decks_card_count_by_entry <- function( entry )
{
  return this.get_sum_of_decks_board( entry.board_type, function( board ) {
    return board.get_count_of( entry.card );
  } );
}

::Organizer.create_first_column_processor <- function( is_japanese )
{
  if ( is_japanese ) {
    return NameJapaneseProcessor();
  }
  return NameProcessor();
}


// -- SubMenu ------------------------------------------------------------
class SubMenu {
  static Separator = null;

  constructor( name, items ) {
    this.name = name;
    this.items = items;
  }

  name = null;
  items = null;
}


// -- j[e` ---------------------------------------------------
// -- Jj[
::Organizer.processors = [
  SubMenu( "O", [
    NameProcessor(),
    NameJapaneseProcessor(),
    ] ),
  BoardTypeProcessor(),
  TotalProcessor(),
  AverageProcessor(),
  SubMenu( "nσfbL", [
    CardCountProcessor( 0 ),
    CardCountProcessor( 1 ),
    CardCountProcessor( 2 ),
    CardCountProcessor( 3 ),
    CardCountProcessor( 4 ),
    ] ),
  SubMenu( "nσfbL", [
    CardCountPercentageProcessor( 0 ),
    CardCountPercentageProcessor( 1 ),
    CardCountPercentageProcessor( 2 ),
    CardCountPercentageProcessor( 3 ),
    CardCountPercentageProcessor( 4 ),
    ] ),
  ];

// -- ёւj[
::Organizer.sorters = [
  StandardSorter(),
  TypeSorter(),
  TotalSorter(),
  ];

// -- ꍀڃXg̏
local special_entries_initialize = function()
{
  ::Organizer.special_entries = [];
  local register = function( board_type ) {
    ::Organizer.special_entries.extend( [
      ::TypeTotalEntry( Card.Type.LandTotal,         Card.Type.Land,         "Land Total",           "ynv",                 board_type ),
      ::TypeTotalEntry( Card.Type.CreatureTotal,     Card.Type.Creature,     "Creature Total",       "N[`[v",         board_type ),
      ::TypeTotalEntry( Card.Type.InstantTotal,      Card.Type.Instant       "Instant Total",        "CX^gv",         board_type ),
      ::TypeTotalEntry( Card.Type.SorceryTotal,      Card.Type.Sorcery       "Socery Total",         "\[T[v",           board_type ),
      ::InstantSorceryTotalEntry( board_type ),
      ::TypeTotalEntry( Card.Type.EnchantmentTotal,  Card.Type.Enchantment,  "Enchantment Total",    "G`gv",         board_type ),
      ::TypeTotalEntry( Card.Type.ArtifactTotal,     Card.Type.Artifact,     "Artifact Total",       "A[eBt@Ngv",     board_type ),
      ::TypeTotalEntry( Card.Type.PlaneswalkerTotal, Card.Type.Planeswalker, "Planeswalker Total",   "vCYEH[J[v", board_type ),
      ::SpellTotalEntry( board_type ),
      ::AllTotalEntry( board_type ),
      ] );
  };
  register( Board.Type.Mainboard );
  ::Organizer.special_entries.append( ::SideboardStartEntry() );
  register( Board.Type.Sideboard );

  ::Organizer.special_entries[0].use  = true; // Land Total
  ::Organizer.special_entries[1].use  = true; // Creature Total
  ::Organizer.special_entries[8].use  = true; // Spell Total
  ::Organizer.special_entries[10].use = true; // Sideboard
}
special_entries_initialize();
